"use strict";
var _a;
Object.defineProperty(exports, "__esModule", { value: true });
exports.Endpoint = exports.RootConfig = void 0;
const tslib_1 = require("tslib");
const _1 = require(".");
const fs_1 = tslib_1.__importDefault(require("fs"));
const ini = require("ini");
// Represents `~/.fauna-shell`
class RootConfig {
    constructor(config) {
        this.defaultEndpoint = config.strOpt("default");
        /**
         * Our updated config uses endpoint.<name> to reserve the endpoint namespace.
         * It is possible prior users still have legacy config that has the endpoints at the top level.
         * When config is updated, we will write the entire file, so it will either
         * all be nested under endpoint or all legacy. The only time this wouldn't be
         * the case is if a user manually modifies their file to have a top level
         * endpoint after we have nested them all under endpoint.  In that scenario,
         * the added endpoint will not be recognized.
         */
        if (RootConfig.configContainsNestedEndpointStructure(config)) {
            this.endpoints = Object.fromEntries(config
                .objectsIn("endpoint")
                .map(([k, v]) => [k, Endpoint.fromConfig(v)]));
        }
        else {
            this.endpoints = Object.fromEntries(config
                .allObjectsWhere((k) => k !== "default")
                .map(([k, v]) => [k, Endpoint.fromConfig(v)]));
        }
        if (this.defaultEndpoint === "default") {
            throw new _1.InvalidConfigError("Default endpoint cannot be named 'default'");
        }
        else if (this.defaultEndpoint &&
            this.endpoints[this.defaultEndpoint] === undefined) {
            throw new _1.InvalidConfigError(`Default endpoint '${this.defaultEndpoint}' was not found`);
        }
    }
    /**
     * If there is an endpoint object in the config, and it has a
     * object underneath it, we are saying that it uses the
     *
     * [endpoint.myacct]
     * secret=***
     *
     * structure. This allows us to support a legacy config that has
     * a top level endpoint named endpoint.  Once we rewrite the config
     * to be endpoint object based adding one manually like that
     * will not be recognized.
     */
    static configContainsNestedEndpointStructure(config) {
        if (config.objectExists("endpoint")) {
            const endpointObj = config.object("endpoint");
            return endpointObj.keys().some((key) => {
                return endpointObj.objectExists(key);
            });
        }
        else {
            return false;
        }
    }
    save(path) {
        const config = this.toIni();
        const encoded = ini.encode(config);
        fs_1.default.writeFileSync(path, encoded);
    }
    toIni() {
        return {
            ...(this.defaultEndpoint !== undefined
                ? { default: this.defaultEndpoint }
                : {}),
            endpoint: Object.fromEntries(Object.entries(this.endpoints).map(([k, v]) => [k, v.toIni()])),
        };
    }
}
exports.RootConfig = RootConfig;
/**
 * Represents an endpoint, or a section of `~/.fauna-shell`.
 *
 * An endpoint contains:
 * - A secret. This is the accout secret for this database.
 * - An optional URL. This is the URL of the database. It defaults to `https://db.fauna.com`.
 * - An optional GraphQL host. This is the host to use for GraphQL requests. It defaults to `graphql.fauna.com`.
 * - An optional GraphQL port. This is the port to use for GraphQL requests. It defaults to `443`.
 */
class Endpoint {
    static fromConfig(config) {
        return new _a({
            secret: config.str("secret"),
            url: _a.getURLFromConfig(config),
            graphqlHost: config.strOpt("graphqlHost"),
            graphqlPort: config.numberOpt("graphqlPort"),
        });
    }
    constructor(opts) {
        var _b, _c, _d;
        this.secret = opts.secret;
        this.url = (_b = opts.url) !== null && _b !== void 0 ? _b : "https://db.fauna.com";
        this.graphqlHost = (_c = opts.graphqlHost) !== null && _c !== void 0 ? _c : "graphql.fauna.com";
        this.graphqlPort = (_d = opts.graphqlPort) !== null && _d !== void 0 ? _d : 443;
    }
    makeScopedEndpoint(scope, role) {
        const secret = this.secret + (scope ? `:${scope}` : "") + (role ? `:${role}` : "");
        return {
            secret,
            url: this.url,
            graphqlHost: this.graphqlHost,
            graphqlPort: this.graphqlPort,
        };
    }
    toIni() {
        return {
            secret: this.secret,
            ...(this.url !== "https://db.fauna.com" ? { url: this.url } : {}),
            ...(this.graphqlHost !== "graphql.fauna.com"
                ? { graphqlHost: this.graphqlHost }
                : {}),
            ...(this.graphqlPort !== 443 ? { graphqlPort: this.graphqlPort } : {}),
        };
    }
}
exports.Endpoint = Endpoint;
_a = Endpoint;
/**
 * Gets a database URL from config.
 */
Endpoint.getURLFromConfig = (config) => {
    return _a.getURL({
        url: config.strOpt("url"),
        scheme: config.strOpt("scheme"),
        domain: config.strOpt("domain"),
        port: config.numberOpt("port"),
    });
};
/**
 * Gets a database URL from command line flags.
 *
 * Note: this is similar to `getURLFromConfig`, but looks up `endpointURL`
 * instead of `url` for the url value.
 */
Endpoint.getURLFromFlags = (flags) => {
    return _a.getURL({
        url: flags.strOpt("endpointURL"),
        scheme: flags.strOpt("scheme"),
        domain: flags.strOpt("domain"),
        port: flags.numberOpt("port"),
    });
};
Endpoint.getURL = (opts) => {
    var _b, _c;
    if (opts.url === undefined &&
        (opts.domain !== undefined ||
            opts.port !== undefined ||
            opts.scheme !== undefined)) {
        const scheme = (_b = opts.scheme) !== null && _b !== void 0 ? _b : "https";
        const domain = (_c = opts.domain) !== null && _c !== void 0 ? _c : "db.fauna.com";
        const port = opts.port ? `:${opts.port}` : "";
        return `${scheme}://${domain}${port}`;
    }
    else {
        return opts.url;
    }
};
